@@ -22,9 +22,9 @@ import simplejson as json
 import sqlalchemy as sa
 import zmq
 
-from sorna.exceptions import (ServiceUnavailable, InvalidAPIParameters, QuotaExceeded,
-                              QueryNotImplemented, InstanceNotFound, KernelNotFound,
-                              SornaError)
+from .exceptions import (ServiceUnavailable, InvalidAPIParameters, QuotaExceeded,
+                         QueryNotImplemented, InstanceNotFound, KernelNotFound,
+                         SornaError)
 from . import GatewayStatus
 from .auth import auth_required
 from .models import KeyPair, Usage
@@ -44,11 +44,12 @@ async def create(request):
     try:
         with _timeout(2):
             params = await request.json()
-        log.info(f"GET_OR_CREATE (lang:{params['lang']}, token:{params['clientSessionToken']})")
+        log.info(f"GET_OR_CREATE (u:{request.keypair['access_key']}, "
+                 f"lang:{params['lang']}, token:{params['clientSessionToken']})")
         assert 8 <= len(params['clientSessionToken']) <= 80
     except (asyncio.TimeoutError, AssertionError,
             KeyError, json.decoder.JSONDecodeError) as e:
-        log.warn(f'GET_OR_CREATE: invalid/missing parameters, {e!r}')
+        log.warning(f'GET_OR_CREATE: invalid/missing parameters, {e!r}')
         raise InvalidAPIParameters
     resp = {}
     try:
@@ -245,6 +246,54 @@ async def instance_stats(app, inst_id, kern_stats, interval):
         await app['registry'].handle_stats(inst_id, kern_stats, interval)
 
 
+async def datadog_update(app):
+    with app['datadog'].statsd as statsd:
+
+        statsd.gauge('sorna.gateway.coroutines', len(asyncio.Task.all_tasks()))
+
+        all_inst_ids = [inst_id async for inst_id in app['registry'].enumerate_instances()]
+        statsd.gauge('sorna.gateway.agent_instances', len(all_inst_ids))
+
+        async with app.dbpool.acquire() as conn, conn.transaction():
+            query = (sa.select([sa.func.sum(KeyPair.c.concurrency_used)])
+                       .select_from(KeyPair))
+            n = await conn.fetchval(query)
+            statsd.gauge('sorna.gateway.active_kernels', n)
+
+            subquery = (sa.select([sa.func.count()])
+                          .select_from(KeyPair)
+                          .where(KeyPair.c.is_active == True)
+                          .group_by(KeyPair.c.user_id))
+            query = sa.select([sa.func.count()]).select_from(subquery.alias())
+            n = await conn.fetchval(query)
+            statsd.gauge('sorna.users.has_active_key', n)
+
+            subquery = subquery.where(KeyPair.c.last_used != None)
+            query = sa.select([sa.func.count()]).select_from(subquery.alias())
+            n = await conn.fetchval(query)
+            statsd.gauge('sorna.users.has_used_key', n)
+
+            query = sa.select([sa.func.count()]).select_from(Usage)
+            n = await conn.fetchval(query)
+            statsd.gauge('sorna.gateway.accum_kernels', n)
+
+
+async def datadog_update_timer(app):
+    if app['datadog'] is None:
+        return
+    while True:
+        try:
+            await datadog_update(app)
+        except asyncio.CancelledError:
+            break
+        except:
+            log.exception('datadog_update unexpected error')
+        try:
+            await asyncio.sleep(5)
+        except asyncio.CancelledError:
+            break
+
+
 async def collect_agent_events(app, heartbeat_interval):
     '''
     Collects agent-generated events for a while (via :func:`grace_event_catcher`).
@@ -317,6 +366,8 @@ async def collect_agent_events(app, heartbeat_interval):
         await update_instance_usage(app, inst_id)
         await app['registry'].forget_instance(inst_id)
 
+    app['kernel_ddtimer'] = asyncio.ensure_future(datadog_update_timer(app))
+
     log.info('entering normal operation mode...')
     app['status'] = GatewayStatus.RUNNING
 
@@ -326,7 +377,7 @@ async def destroy(request):
     if request.app['status'] != GatewayStatus.RUNNING:
         raise ServiceUnavailable('Server not ready.')
     kern_id = request.match_info['kernel_id']
-    log.info(f'DESTROY (k:{kern_id})')
+    log.info(f"DESTROY (u:{request.keypair['access_key']}, k:{kern_id})")
     try:
         await request.app['registry'].destroy_kernel(kern_id)
     except SornaError:
@@ -341,7 +392,7 @@ async def get_info(request):
         raise ServiceUnavailable('Server not ready.')
     resp = {}
     kern_id = request.match_info['kernel_id']
-    log.info(f'GETINFO (k:{kern_id})')
+    log.info(f"GETINFO (u:{request.keypair['access_key']}, k:{kern_id})")
     try:
         kern = await request.app['registry'].get_kernel(kern_id)
         await request.app['registry'].update_kernel(kern_id, {
@@ -374,7 +425,7 @@ async def restart(request):
     if request.app['status'] != GatewayStatus.RUNNING:
         raise ServiceUnavailable('Server not ready.')
     kern_id = request.match_info['kernel_id']
-    log.info(f'RESTART (k:{kern_id})')
+    log.info(f"RESTART (u:{request.keypair['access_key']}, k:{kern_id})")
     try:
         kern = await request.app['registry'].get_kernel(kern_id)
         await request.app['registry'].update_kernel(kern_id, {
@@ -401,9 +452,9 @@ async def execute_snippet(request):
     try:
         with _timeout(2):
             params = await request.json()
-        log.info(f'EXECUTE_SNIPPET (k:{kern_id})')
+        log.info(f"EXECUTE_SNIPPET (u:{request.keypair['access_key']}, k:{kern_id})")
     except (asyncio.TimeoutError, json.decoder.JSONDecodeError):
-        log.warn('EXECUTE_SNIPPET: invalid/missing parameters')
+        log.warning('EXECUTE_SNIPPET: invalid/missing parameters')
         raise InvalidAPIParameters
     try:
         kern = await request.app['registry'].get_kernel(kern_id)
@@ -583,6 +634,9 @@ async def init(app):
 
 
 async def shutdown(app):
+    if 'kernel_ddtimer' in app and not app['kernel_ddtimer'].done():
+        app['kernel_ddtimer'].cancel()
+        await app['kernel_ddtimer']
     for per_kernel_handlers in app['stream_pty_handlers'].values():
         for handler in per_kernel_handlers.copy():
             handler.cancel()
