@@ -13,10 +13,15 @@ import aiohttp
 from aiohttp import web
 import asyncpgsa
 import uvloop
+try:
+    import datadog
+    datadog_available = True
+except ImportError:
+    datadog_available = False
 
 from sorna.argparse import ipaddr, path, port_no
-from sorna.exceptions import (SornaError, GenericNotFound,
-                              GenericBadRequest, InternalServerError)
+from .exceptions import (SornaError, GenericNotFound,
+                         GenericBadRequest, InternalServerError)
 from sorna.utils import env_info
 from ..manager import __version__
 from . import GatewayStatus
@@ -25,6 +30,7 @@ from .config import load_config, init_logger
 from .events import init as event_init, shutdown as event_shutdown
 from .kernel import init as kernel_init, shutdown as kernel_shutdown
 from .ratelimit import init as rlim_init, shutdown as rlim_shutdown
+from .utils import prettify_traceback
 
 LATEST_API_VERSION = 'v1.20160915'
 
@@ -45,17 +51,40 @@ async def on_prepare(request, response):
 async def exception_middleware_factory(app, handler):
     async def exception_middleware_handler(request):
         try:
-            return (await handler(request))
-        except SornaError:
+            if app['datadog']:
+                app['datadog'].statsd.increment('sorna.gateway.api.requests')
+            resp = (await handler(request))
+        except SornaError as ex:
+            if app['datadog']:
+                statsd = app['datadog'].statsd
+                statsd.increment('sorna.gateway.api.failures')
+                statsd.increment(f'sorna.gateway.api.status.{ex.status_code}')
             raise
         except web.HTTPException as ex:
+            if app['datadog']:
+                statsd = app['datadog'].statsd
+                statsd.increment('sorna.gateway.api.failures')
+                statsd.increment(f'sorna.gateway.api.status.{ex.status_code}')
             if ex.status_code == 404:
                 raise GenericNotFound
             log.warning(f'Bad request: {ex!r}')
             raise GenericBadRequest
         except Exception as ex:
             log.exception('Uncaught exception in HTTP request handlers')
+            title = f'Exception from {request.method} {request.rel_url.path}'
+            tag = f'path:{request.rel_url.path}'
+            text = prettify_traceback(ex)
+            if app['datadog']:
+                app['datadog'].statsd.event(
+                    title, text,
+                    tags=['sorna', 'exception'],
+                    aggregation_key=request.rel_url.path,
+                    alert_type='error')
             raise InternalServerError
+        else:
+            if app['datadog']:
+                app['datadog'].statsd.increment(f'sorna.gateway.api.status.{resp.status}')
+            return resp
     return exception_middleware_handler
 
 
@@ -63,6 +92,16 @@ async def gw_init(app):
     app.on_response_prepare.append(on_prepare)
     app.router.add_route('GET', '/v1', hello)
     app['status'] = GatewayStatus.STARTING
+    app['datadog'] = None
+    if datadog_available:
+        if app.config.datadog_api_key is None:
+            log.info('skipping datadog initialization due to missing API key...')
+        else:
+            datadog.initialize(
+                api_key=app.config.datadog_api_key,
+                app_key=app.config.datadog_app_key)
+            app['datadog'] = datadog
+            log.info('datadog logging enabled.')
 
     app.dbpool = await asyncpgsa.create_pool(
         host=str(app.config.db_addr[0]),
@@ -95,6 +134,11 @@ def gw_args(parser):
     parser.add('--ssl-key', env_var='SORNA_SSL_KEY', type=path, default=None,
                help='The path to the private key used to make requests for the SSL certificate. '
                     '(default: None)')
+    if datadog_available:
+        parser.add('--datadog-api-key', env_var='DATADOG_API_KEY', type=str, default=None,
+                   help='The API key for Datadog monitoring agent.')
+        parser.add('--datadog-app-key', env_var='DATADOG_APP_KEY', type=str, default=None,
+                   help='The application key for Datadog monitoring agent.')
 
 
 def main():
